#------------------------------------------------------------------
# $Copyright: Copyright (c) 2016 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $Header: /project/isv-agent/VProAgents/WebSphereMQFTE/WebSphereMQFTE.pm,v 1.8 2010/02/23 05:39:08 rkumar Exp $
#
# $Author: rkumar $
# $Date: 2010/02/23 05:39:08 $
# $Revision: 1.8 $
#
# Provides:                            WebSphereMQFT
# Author:                              Rajeev Kumar
# crAgent Build:                       1.88 2008/10/13 08:28:02
# crAgent Run Date:                    14 Apr, 2009
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package WebSphereMQFTE;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use File::Path qw ( mkpath );
use File::stat;
use POSIX qw ( uname );

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my @liRunningProcs = (); # Global..

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: WebSphereMQFTE.pm,v 1.8 2010/02/23 05:39:08 rkumar Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#------------------------------------
# Define WebSphereMQFTE subroutines..
#------------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String FTEAgentName,
                              IN String CoordinationQM,
                              IN String AgentQM,
                              IN String HomeDir,
                              IN String DataDir,
                              IN String FTEUser,
                              IN String EnvFile,
                              IN String MonitorProgram,
                              IN Integer SecondLevelMonitor );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName           Entry pt name
 $localFTEAgentName         Name of WebSphere MQ FTE agent
 $localCoordinationQM       Coordination Queue Manager of the WebSphere MQ FTE agent
 $localAgentQM              Agent Queue Manager of the WebSphere MQ FTE agent
 $localHomeDir              Home or Installation directory of the WebSphere MQ FTE
 $localDataDir              Data Directory of the WebSphere MQ FTE
 $localFTEUser              Unix user owner for WebSphere MQ FTE
 $localEnvFile              Absolute path to environment file to source
 $localMonitorProgram       Full path to external monitor program
 $localSecondLevelMonitor   Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localFTEAgentName,
       $localCoordinationQM,
       $localAgentQM,
       $localHomeDir,
       $localDataDir,
       $localFTEUser,
       $localEnvFile,
       $localMonitorProgram,
       $localSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20017, "Subroutine <%s> called with:
  EntryPointName           [%s]
  FTEAgentName             [%s]
  CoordinationQM           [%s]
  AgentQM                  [%s]
  HomeDir                  [%s]
  DataDir                  [%s]
  FTEUser                  [%s]
  EnvFile                  [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,          $sEntryPointName,          $localFTEAgentName,
  $localCoordinationQM,        $localAgentQM,        $localHomeDir,
  $localDataDir,           $localFTEUser,            $localEnvFile,
  $localMonitorProgram, $localSecondLevelMonitor );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName  FTEAgentName   CoordinationQM   AgentQM   HomeDir
       DataDir FTEUser EnvFile MonitorProgram SecondLevelMonitor );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( FTEAgentName CoordinationQM FTEUser HomeDir );
  my @lsOfflineArgs = qw ( FTEAgentName CoordinationQM FTEUser HomeDir AgentQM );
  my @lsMonitorArgs = qw ( FTEAgentName CoordinationQM FTEUser DataDir );
  my @lsCleanArgs   = qw ( FTEAgentName CoordinationQM FTEUser DataDir );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20018,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  
  #---------------------------------------------------
  # Validation for online and offline entry point(s)..
  #---------------------------------------------------
  if ( $sEntryPointName eq "online"
       || $sEntryPointName eq "offline" ) {
   
    #-----------------------------
    # Validate parameter FTEUser..
    #-----------------------------
    if ( ! defined getpwnam $localFTEUser ) {
      &logError ( 20019, "<%s> FTEUser [%s] is not a valid UNIX login",
      $sMyName, $localFTEUser );
      $iReturnCode = 0; # Failure..
    }

    #-------------------
    # Validate EnvFile..
    #-------------------
    if ( $localEnvFile =~ /\S/ && ! -e $localEnvFile ) {
      &logError ( 20032,
      "<%s> EnvFile [%s] does not exist",
      $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }

    #-----------------------------------------
    # Validate parameter HomeDir and DataDir..
    #-----------------------------------------
    if ( ! -d $localHomeDir) {
      &logError ( 20020, "<%s> Home Directory [%s] do not exists",
       $sMyName, $localHomeDir );
       $iReturnCode = 0; # Failure..
    }
    if ( ! -d $localDataDir) {
      &logError ( 20021, "<%s> Data Directory [%s] do not exists",
       $sMyName, $localDataDir );
       $iReturnCode = 0; # Failure..
    }
  }
  
  #----------------------------------
  # Validation for all entry points..
  #----------------------------------

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) { 
    #---------------------------
    # Validate Agent Directory..
    #---------------------------
    my $localAgentDir =
    "$localDataDir/config/$localCoordinationQM/agents/$localFTEAgentName";

    if( ! -d $localAgentDir ) {
       &logError ( 20022,
       "<%s>  Agent Directory [%s] for FTEAgent [%s] do not exists",
       $sMyName, $localAgentDir, $localFTEAgentName );
       $iReturnCode = 0; # Failure..
    }

    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20023, "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }
  }

  $iReturnCode ||
  &logWarn ( 20024,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String AgentQM,
                         IN String CoordinationQM,
                         IN String DataDir,
                         IN String EnvFile,
                         IN String FTEAgentName,
                         IN String FTEUser,
                         IN String HomeDir,
                         IN Integer SecondLevelMonitor );

 Description : Start logic needed for online entry pt

 Parameters :

 $sAgentQM              Agent Queue Manager of the WebSphere MQ FTE agent
 $sCoordinationQM       Coordination Queue Manager of the WebSphere MQ FTE agent
 $sDataDir              Data Directory of the WebSphere MQ FTE
 $sEnvFile              Absolute path to environment file to source
 $sFTEAgentName         Name of WebSphere MQ FTE agent
 $sFTEUser              Unix user owner for WebSphere MQ FTE
 $sHomeDir              Home or Installation directory of the WebSphere MQ FTE
 $iSecondLevelMonitor   Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAgentQM,
       $sCoordinationQM,
       $sDataDir,
       $sEnvFile,
       $sFTEAgentName,
       $sFTEUser,
       $sHomeDir,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20025, "Subroutine <%s> called with:
  AgentQM                  [%s]
  CoordinationQM           [%s]
  DataDir                  [%s]
  EnvFile                  [%s]
  FTEAgentName             [%s]
  FTEUser                  [%s]
  HomeDir                  [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,  $sAgentQM,  $sCoordinationQM,  $sDataDir,   $sEnvFile,
  $sFTEAgentName, $sFTEUser, $sHomeDir, $iSecondLevelMonitor );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #-------------------------------------------------
  # Whack any preexisting WebSphereMQFTE processes..
  #-------------------------------------------------
  &logTrace ( 20026,
  "<%s> Killing WebSphereMQFTE processes for this instance", $sMyName );

  if ( WebSphereMQFTE::KillWebSphereMQFTEProcesses ( $sCoordinationQM,
                                                     $sDataDir,
                                                     $sFTEAgentName,
                                                     $sFTEUser ) ) {
    &logTrace ( 20027,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20028,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
  
  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + WebSphereMQFTE::GetEngineTimeout ( 1 );
  
  #------------------------------------------
  # Define start command and args variables..
  #------------------------------------------
  $sCommand = "$sHomeDir/bin/fteStartAgent";
  $sCommandArgs = "-p $sCoordinationQM $sFTEAgentName";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20029,
  "<%s> Starting WebSphereMQFTE instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20030,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    WebSphereMQFTE::ExecLocalCmd ( $sFTEUser, 
                                   $sEnvFile, 
                                   $sCommand, 
                                   $sCommandArgs, 
                                   \$iExitCode, 
                                   $iTimeout, 
                                   \@lsCmdout ) ) {
    &logTrace ( 20031,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode ;
    
  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20032,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         WebSphereMQFTE::GetProcState ( \@liPids,
                                         $sCoordinationQM,
                                         $sDataDir,
                                         $sFTEAgentName,
                                         $sFTEUser,
                                         1 ) && (        # Running FFLM..
         $iSecondLevelMonitor  ?                         # If enabled..
         VCSDef_MonitorExit::ONLINE ==                   # Online after..
         WebSphereMQFTE::SecondLevelMonitor ( $sAgentQM, # Running SLM..
                                              $sCoordinationQM,
                                              $sEnvFile,
                                              $sFTEAgentName,
                                              $sFTEUser,
                                              $sHomeDir,
                                              $sDataDir,
                                              $iTillWhen ) : 1 ) ) { 
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20033,
      "<%s> Confirmed start-up of this instance of WebSphereMQFTE server",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logError ( 20034,
      "<%s> WebSPhereMQFTE instance NOT online after SLM. ",
      $sMyName );
      $iReturnCode = 0; # Failure..
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String AgentQM,
                        IN String CoordinationQM,
                        IN String DataDir,
                        IN String EnvFile,
                        IN String FTEAgentName,
                        IN String FTEUser,
                        IN String HomeDir );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sAgentQM          Agent Queue Manager of the WebSphere MQ FTE agent
 $sCoordinationQM   Coordination Queue Manager of the WebSphere MQ FTE agent
 $sDataDir          Data Directory of the WebSphere MQ FTE
 $sEnvFile          Absolute path to environment file to source
 $sFTEAgentName     Name of WebSphere MQ FTE agent
 $sFTEUser          Unix user owner for WebSphere MQ FTE
 $sHomeDir          Home or Installation directory of the WebSphere MQ FTE
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAgentQM,
       $sCoordinationQM,
       $sDataDir,
       $sEnvFile,
       $sFTEAgentName,
       $sFTEUser,
       $sHomeDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20035, "Subroutine <%s> called with:
  AgentQM                  [%s]
  CoordinationQM           [%s]
  DataDir                  [%s]
  EnvFile                  [%s]
  FTEAgentName             [%s]
  FTEUser                  [%s]
  HomeDir                  [%s]",

  $sMyName,  $sAgentQM,  $sCoordinationQM,  $sDataDir,   $sEnvFile,
  $sFTEAgentName, $sFTEUser, $sHomeDir );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + WebSphereMQFTE::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
  
  #------------------------------------
  # Define stop command and arguments..
  #------------------------------------
  $sCommand = "$sHomeDir/bin/fteStopAgent";
  $sCommandArgs = "-m $sAgentQM -p $sCoordinationQM $sFTEAgentName";
  ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 

  &logTrace ( 20036,
  "<%s> Stopping WebSphereMQFTE with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #-------------------
  # Request shutdown..
  #-------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20037,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          WebSphereMQFTE::ExecLocalCmd ( $sFTEUser, 
                                         $sEnvFile, 
                                         $sCommand, 
                                         $sCommandArgs, 
                                         \$iExitCode, 
                                         $iTimeout, 
                                         \@lsCmdout ) ) {
    &logTrace ( 20038,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #--------------------
  # Delay entry point..
  #--------------------
  if ( $iReturnCode ) { # Success..
    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20039,
    "<%s> Delaying exit to confirm shutdown..", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      WebSphereMQFTE::GetProcState ( \@liPids,
                                     $sCoordinationQM,
                                     $sDataDir,
                                     $sFTEAgentName,
                                     $sFTEUser,
                                     0 ); # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        &logTrace ( 20040,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout - $iSleepTime );
      }
      else {
        &logInfo ( 20041,
        "<%s> Confirmed shutdown of WebSphereMQFTE instance",
        $sMyName );
        last; # Get out..
      }
    }
  }

  #------------------------------------------------
  # Whack any persisting WebSphereMQFTE processes..
  #------------------------------------------------
  &logTrace ( 20042,
  "<%s> Killing WebSphereMQFTE processes for this instance", $sMyName );

  if ($iReturnCode =
      WebSphereMQFTE::KillWebSphereMQFTEProcesses ( $sCoordinationQM,
                                                    $sDataDir,
                                                    $sFTEAgentName,
                                                    $sFTEUser ) ) {
    &logTrace ( 20043,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20044,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN String AgentQM,
                        IN Integer CleanReason,
                        IN String CoordinationQM,
                        IN String DataDir,
                        IN String EnvFile,
                        IN String FTEAgentName,
                        IN String FTEUser,
                        IN String HomeDir );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $sAgentQM                  Agent Queue Manager of the WebSphere MQ FTE agent
 $iCleanReason              Why clean was called
 $sCoordinationQM           Coordination Queue Manager of the WebSphere MQ FTE agent
 $sDataDir                  Data Directory of the WebSphere MQ FTE
 $sEnvFile                  Absolute path to environment file to source
 $sFTEAgentName             Name of WebSphere MQ FTE agent
 $sFTEUser                  Unix user owner for WebSphere MQ FTE
 $sHomeDir                  Home or Installation directory of the WebSphere MQ FTE
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAgentQM,
       $iCleanReason,
       $sCoordinationQM,
       $sDataDir,
       $sEnvFile,
       $sFTEAgentName,
       $sFTEUser,
       $sHomeDir ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20045, "Subroutine <%s> called with:
  AgentQM                  [%s]
  CleanReason              [%s]
  CoordinationQM           [%s]
  DataDir                  [%s]
  EnvFile                  [%s]
  FTEAgentName             [%s]
  FTEUser                  [%s]
  HomeDir                  [%s]",

  $sMyName, $sAgentQM, $sCleanReason, $sCoordinationQM,  $sDataDir,
  $sEnvFile, $sFTEAgentName, $sFTEUser, $sHomeDir );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + WebSphereMQFTE::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsFiles, @lsCmdout ) = ();
  
  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  $sCommand = "$sHomeDir/bin/fteStopAgent";
  $sCommandArgs = "-m $sAgentQM -p $sCoordinationQM -i $sFTEAgentName";

  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    &logInfo ( 20046,
    "<%s> Attempting to stop WebSphereMQFTE with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #-------------------
    # Request shutdown..
    #-------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20047,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( WebSphereMQFTE::ExecLocalCmd ( $sFTEUser, 
                                           $sEnvFile, 
                                           $sCommand, 
                                           $sCommandArgs, 
                                           \$iExitCode, 
                                           $iTimeout, 
                                           \@lsCmdout )) {
      &logTrace ( 20048, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }

  #------------------------------------------------
  # Whack any persisting WebSphereMQFTE processes..
  #------------------------------------------------
  &logTrace ( 20049,
  "<%s> Killing WebSphereMQFTE processes for this instance", $sMyName );

  if ( WebSphereMQFTE::KillWebSphereMQFTEProcesses ( $sCoordinationQM,
                                                     $sDataDir,
                                                     $sFTEAgentName,
                                                     $sFTEUser ) ) {
    &logTrace ( 20050,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20051,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String CoordinationQM,
                                   IN String DataDir,
                                   IN String FTEAgentName,
                                   IN String FTEUser );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sCoordinationQM           Coordination Queue Manager of the WebSphere MQ FTE agent
 $sDataDir                  Data Directory of the WebSphere MQ FTE
 $sFTEAgentName             Name of WebSphere MQ FTE agent
 $sFTEUser                  Unix user owner for WebSphere MQ FTE
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCoordinationQM,
       $sDataDir,
       $sFTEAgentName,
       $sFTEUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20052, "Subroutine <%s> called:
  CoordinationQM           [%s]
  DataDir                  [%s]
  FTEAgentName             [%s]
  FTEUser                  [%s]",

  $sMyName, $sCoordinationQM, $sDataDir, $sFTEAgentName, $sFTEUser );

  my @liDummyPids = ();
  my $iReturnCode = WebSphereMQFTE::GetProcState ( \@liDummyPids,
                                                   $sCoordinationQM,
                                                   $sDataDir,
                                                   $sFTEAgentName,
                                                   $sFTEUser,
                                                   1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20053,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20054,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20055,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String CoordinationQM,
                               IN String DataDir,
                               IN String FTEAgentName,
                               IN String FTEUser,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sCoordinationQM           Coordination Queue Manager of the WebSphere MQ FTE agent
 $sDataDir                  Data Directory of the WebSphere MQ FTE
 $sFTEAgentName             Name of WebSphere MQ FTE agent
 $sFTEUser                  Unix user owner for WebSphere MQ FTE
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sCoordinationQM,
       $sDataDir,
       $sFTEAgentName,
       $sFTEUser,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20056, "Subroutine <%s> called with:
  OnlinePids               [%s]
  CoordinationQM           [%s]
  DataDir                  [%s]
  FTEAgentName             [%s]
  FTEUser                  [%s]
  UseFFLM                  [%s]",

  $sMyName,    $rliOnlinePids,     $sCoordinationQM,     $sDataDir,
  $sFTEAgentName, $sFTEUser, $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues,
       %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  @$rliOnlinePids = ();

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( WebSphereMQFTE::GetSrvrDetails ( $sCoordinationQM, 
                                          $sDataDir, 
                                          $sFTEAgentName, 
                                          \%hsSrvrDetails ) ) {
      &logTrace ( 20057,
      "<%s> Got details for monitoring WebSphereMQFTE instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20058,
      "<%s> Failed to get details for monitoring WebSphereMQFTE instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nWebSphereMQFTE process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = WebSphereMQFTE::FFLMStatus ( $sPidFile,
                                                 $sProc,
                                                 \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20059,
      "<%s> WebSphereMQFTE instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20060,
    "<%s> Called to get processes for User [%s] ",#UID [%s]",
    $sMyName, $sFTEUser);    # $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20061,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20062,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20063,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20064, "<%s> Searching UID [%s] PID [%s] PPID [%s] CMD [%s]",
      $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
      $$rhlProc { COMMAND } );

      #-----------------------------------------------
      # Look for processes that were offline in FFLM..
      #-----------------------------------------------
      if ( $iUseFFLM != 2 ) { # Request to filter..
        foreach $sKey ( keys %hsCopy ) {
          @lsUpdates = (); # Reset..
          @lsValues = @{ $hsCopy { $sKey } }; # Get values..
  
          #--------------------------------------------------
          # Should have details per three elements of array..
          # <PIDFILE> <PROCESS> <STATE>
          #--------------------------------------------------
          while ( ( $sPidFile, $sProc, $iFound ) =
                    splice ( @lsValues, 0, 3 ) ) {
            if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
              if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {    # Now found..
                unshift ( @$rliOnlinePids, $iPid );         # Store..
  
                &logTrace ( 20065,
                "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                $sProc );

                $hiProcPids { $sKey } = $iPid;        # Store..
                $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
              }
            }
  
            #----------------
            # Update status..
            #----------------
            push ( @lsUpdates, $sPidFile, $sProc, $iFound );
          }
  
          @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
        }
      }
      else { # Request for kill..
        &logTrace ( 20066,
        "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
        $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
        push ( @$rliOnlinePids, $iPid ); # Store..
      } 
    } 
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20067,
    "<%s> Found [%s] PIDs [%s] for this instance of WebSphereMQFTE.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20068,
    "<%s> WebSphereMQFTE process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String AgentQM,
                          IN String CoordinationQM,
                          IN String EnvFile,
                          IN String FTEAgentName,
                          IN String FTEUser,
                          IN String HomeDir,
                          IN String DataDir,
                          IN String MonitorProgram,
                          IN Integer SecondLevelMonitor );

 Description: Performs the in-depth monitoring

 Parameters :

 $sAgentQM                  Agent Queue Manager of the WebSphere MQ FTE agent
 $sCoordinationQM           Coordination Queue Manager of the WebSphere MQ FTE agent
 $sEnvFile                  Absolute path to environment file to source
 $sFTEAgentName             Name of WebSphere MQ FTE agent
 $sFTEUser                  Unix user owner for WebSphere MQ FTE
 $sHomeDir                  Home or Installation directory of the WebSphere MQ FTE
 $sDataDir                  Data Directory of the WebSphere MQ FTE
 $sMonitorProgram           Full path to external monitor program
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAgentQM,
       $sCoordinationQM,
       $sEnvFile,
       $sFTEAgentName,
       $sFTEUser,
       $sHomeDir,
       $sDataDir,
       $sMonitorProgram,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20069, "Subroutine <%s> called with:
  AgentQM                  [%s]
  CoordinationQM           [%s]
  EnvFile                  [%s]
  FTEAgentName             [%s]
  FTEUser                  [%s]
  HomeDir                  [%s]
  DataDir                  [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]",

  $sMyName, $sAgentQM, $sCoordinationQM, $sEnvFile, $sFTEAgentName,
  $sFTEUser, $sHomeDir, $sDataDir,$sMonitorProgram, $iSecondLevelMonitor );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         WebSphereMQFTE::DelayedAttributeCheck ( $sAgentQM,
                                                 $sCoordinationQM,
                                                 $sEnvFile,
                                                 $sFTEAgentName,
                                                 $sFTEUser,
                                                 $sHomeDir ) ) ) {
      &logTrace ( 20070,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20071,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20072,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + WebSphereMQFTE::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         WebSphereMQFTE::SecondLevelMonitor ( $sAgentQM,
                                              $sCoordinationQM,
                                              $sEnvFile,
                                              $sFTEAgentName,
                                              $sFTEUser,
                                              $sHomeDir,
                                              $sDataDir,
                                              $iTillWhen ) ) ) {

      &logTrace ( 20073,
      "<%s> WebSphereMQFTE instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20074,
      "<%s> WebSphereMQFTE instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + WebSphereMQFTE::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         WebSphereMQFTE::ExternalMonitor ( $sFTEUser,
                                           $sEnvFile,
                                           $sExtCmd,
                                           $sExtCmdArgs,
                                           $iTillWhen ) ) ) {
      &logTrace ( 20075,
      "<%s> WebSphereMQFTE instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20076,
      "<%s> WebSphereMQFTE instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String AgentQM,
                                        IN String CoordinationQM,
                                        IN String EnvFile,
                                        IN String FTEAgentName,
                                        IN String FTEUser,
                                        IN String HomeDir );

 Description : Just In Time Validation

 Parameters :

 $sAgentQM                  Agent Queue Manager of the WebSphere MQ FTE agent
 $sCoordinationQM           Coordination Queue Manager of the WebSphere MQ FTE agent
 $sEnvFile                  Absolute path to environment file to source
 $sFTEAgentName             Name of WebSphere MQ FTE agent
 $sFTEUser                  Unix user owner for WebSphere MQ FTE
 $sHomeDir                  Home or Installation directory of the WebSphere MQ FTE
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAgentQM,
       $sCoordinationQM,
       $sEnvFile,
       $sFTEAgentName,
       $sFTEUser,
       $sHomeDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20077, "Subroutine <%s> called with:
  AgentQM                  [%s]
  CoordinationQM           [%s]
  EnvFile                  [%s]
  FTEAgentName             [%s]
  FTEUser                  [%s]
  HomeDir                  [%s]",

  $sMyName, $sAgentQM, $sCoordinationQM, $sEnvFile, $sFTEAgentName,
  $sFTEUser, $sHomeDir );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( AgentQM CoordinationQM EnvFile FTEAgentName FTEUser HomeDir );
  my @lsReqdArgs = qw ( AgentQM CoordinationQM FTEAgentName FTEUser HomeDir );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20078,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-------------------------------
  # Individual validation begins..
  #-------------------------------
  if ( $sEnvFile =~ /\S/  && ! -e $sEnvFile ) {
    &logError ( 20078,
    "<%s> EnvFile [%s] do not exist",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  if ( ! -d $sHomeDir) {
   &logError ( 20020, "<%s> Home Directory [%s] do not exists",
   $sMyName, $sHomeDir );
   $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String AgentQM,
                                     IN String CoordinationQM,
                                     IN String EnvFile,
                                     IN String FTEAgentName,
                                     IN String FTEUser,
                                     IN String HomeDir,
                                     IN String DataDir,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sAgentQM                  Agent Queue Manager of the WebSphere MQ FTE agent
 $sCoordinationQM           Coordination Queue Manager of the WebSphere MQ FTE agent
 $sEnvFile                  Absolute path to environment file to source
 $sFTEAgentName             Name of WebSphere MQ FTE agent
 $sFTEUser                  Unix user owner for WebSphere MQ FTE
 $sHomeDir                  Home or Installation directory of the WebSphere MQ FTE
 $sDataDir                  Data Directory of the WebSphere MQ FTE
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAgentQM,
       $sCoordinationQM,
       $sEnvFile,
       $sFTEAgentName,
       $sFTEUser,
       $sHomeDir,
       $sDataDir,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20079, "Subroutine <%s> called with:
  AgentQM                  [%s]
  CoordinationQM           [%s]
  EnvFile                  [%s]
  FTEAgentName             [%s]
  FTEUser                  [%s]
  HomeDir                  [%s]
  DataDir                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sAgentQM, $sCoordinationQM, $sEnvFile, $sFTEAgentName,
  $sFTEUser, $sHomeDir, $sDataDir, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;
  
  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout,
        $sSourceFile, $sDesFile, $sDesDir ) = ();

  my $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME }; 
  
  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  $sSourceFile =  "$sDataDir/config/$sCoordinationQM/agents/" . 
                  "$sFTEAgentName/agent.properties";
  $sDesFile     = "/tmp/.$sResourceName" . "_VCS_WMQFTE.slm";
  $sCommand     = "$sHomeDir/bin/fteCreateTransfer";
  $sCommandArgs = "-sa $sFTEAgentName -sm $sAgentQM -da $sFTEAgentName" .
                  " -dm $sAgentQM -df $sDesFile  -pr 0 -p $sCoordinationQM" .
                  " -cs MD5 $sSourceFile";

  #------------------------------------------------
  # Validate and Create Destination file Directory.
  #------------------------------------------------
  $sDesDir  =  "/tmp/.VRTSWebSphereMQFTE";

  if ( -d $sDesDir ) {
    &logTrace ( 20080, "<%s> Destination folder [%s] already exists",
    $sMyName, $sDesDir );
  }
  elsif ( ! mkpath ( $sDesDir, 0, 0700 ) ) {
    &logWarn ( 20081, "<%s> Failed to create Destination folder [%s]",
    $sMyName, $sDesDir );
    return VCSDef_MonitorExit::UNKNOWN; # Cannot proceed..
  }
  
  #------------------------------------
  # Destination file should not exist..
  #------------------------------------
  if ( -f $sDesFile && ! Sys::UnlinkFilesAsUser ( $sDesFile ) ) {
    &logError ( 20082,
    "<%s> Could not remove file [%s]", $sMyName, $sDesFile );
    return VCSDef_MonitorExit::UNKNOWN; # Cannot proceed..
  }

  #------------------------
  # Check the source File..
  #------------------------
  if( ! -f $sSourceFile){
    &logError ( 20065,
    "<%s> FTE Agent Properties file [%s] do not exist", 
    $sMyName, $sFTEAgentName );
    return VCSDef_MonitorExit::UNKNOWN;
  }
  
  #--------------
  #Set LANG = C..
  #--------------
  my $sEnv_before = $ENV { LANG };
  $ENV { LANG } = "C" ;
  
  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20083,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( WebSphereMQFTE::ExecLocalCmd ( $sFTEUser, 
                                         $sEnvFile, 
                                         $sCommand, 
                                         $sCommandArgs, 
                                         \$iExitCode, 
                                         $iTimeout, 
                                         \@lsCmdout ) ) {
    &logTrace ( 20084,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
       &logError ( 20085,
       "<%s> Insufficient time to complete in-depth monitoring.
       Short by [%s] sec",
       $sMyName, abs ( $iTimeout ) );

      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    elsif ( WebSphereMQFTE::CheckFileTransfer( $sFTEUser,
                                               $sEnvFile,
                                               $sFTEAgentName,
                                               $sAgentQM,
                                               $sCoordinationQM,
                                               $sHomeDir,
                                               \@lsCmdout,
                                               $iTillWhen ) ) {
         &logTrace ( 20086,
         "<%s> File [%s] is transfered successfully to Destination [%s]",
         $sMyName, $sSourceFile, $sDesFile );
    }
    else {
         $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
  }
  elsif ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #-------------------------------------
  # Re-set LANG to the original value..
  #-------------------------------------

  defined $sEnv_before ? $ENV { LANG } = $sEnv_before : delete $ENV { LANG };

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int CheckFileTransfer ( IN String FTEUser,
                                    IN String EnvFile,
                                    IN String FTEAgentName,
                                    IN String AgentQM,
                                    IN String CoordinationQM,
                                    IN String HomeDir,
                                    IN ListReference Cmdout,
                                    IN Integer TillWhen );

 Description : Check whether file is got transfered successfully or not. 
               If not just cancel the request transfer.

 Parameters :

 $sFTEUser            Unix user owner for WebSphere MQ FTE
 $sEnvFile            Absolute path to environment file to source
 $sFTEAgentName       Name of WebSphere MQ FTE agent
 $sAgentQM            Agent Queue Manager of the WebSphere MQ FTE agent
 $sCoordinationQM     Coordination Queue Manager of the WebSphere MQ FTE agent
 $sHomeDir            Home or Installation directory of the WebSphere MQ FTE
 $rlsCmdout           List Reference for the output of the command fteCreateTransfer
 $iTillWhen           Time in EPOCH left for IInd level monitoring

 Returns  : Integer value as defined below:

 1: If the file successfully got transfered
 0: If file not get transfered
 -------------------------------------------------------------------------

=cut

sub CheckFileTransfer {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sFTEUser,
       $sEnvFile,
       $sFTEAgentName,
       $sAgentQM,
       $sCoordinationQM,
       $sHomeDir,
       $rlsCmdout,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20087, "Subroutine <%s> called with:
  FTEUser          [%s]
  EnvFile          [%s]
  FTEAgentName     [%s]
  AgentQM          [%s]
  CoordinationQM   [%s]
  HomeDir          [%s]
  rlsCmdout        [%s]
  TillWhen         [%s]",

  $sMyName, $sFTEUser, $sEnvFile, $sFTEAgentName, $sAgentQM,
  $sCoordinationQM, $sHomeDir, $rlsCmdout, $iTillWhen );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;
  
  #-------------------------
  # Define local variables..
  #-------------------------
  my $iSleepTime = 10; # Secs..
  my ( $iTimeout, $sTransferId, $sResourceName, $sDesFile,
       $sStr, $sCommand, $sCommandArgs, $iExitCode, @lsCmdout ) = ();

  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sDesFile = "/tmp/.$sResourceName"."_VCS_WMQFTE.slm";

  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    #--------------------------------------
    # Unlink temporary file we transfered..
    #--------------------------------------
    if ( -f $sDesFile ) {
      unless ( Sys::UnlinkFilesAsUser ( $sDesFile ) ) {
        &logError ( 20088,
        "<%s> Could not remove file [%s]", $sMyName, $sDesFile );
        return 1; # Failure..
      }
      else {
        &logTrace ( 20089,
        "<%s> Deleted file [%s]. Returning [%s]",
        $sMyName, $sDesFile, $iReturnCode );
        return $iReturnCode;
      }
    }
    else { # Wait for file to come..
      &logTrace ( 20090,
      "<%s> Confirming file transfer with [%s] sec in hand",
      $sMyName, $iTimeout - $iSleepTime );

      sleep $iSleepTime;
    }
  }

  #-------------------------------
  # If we are hare, we've failed..
  #-------------------------------

  #---------------------------------
  # Extract Transaction request ID..
  #---------------------------------
  foreach $sStr ( @$rlsCmdout ) {
    if ( $sStr =~ /(.*)The request ID is:\s(.*)/ ) {
      $sTransferId = $2;
      last;
    }
  }

  #----------------------------
  # Cancel Requested Transfer..
  #----------------------------
  $sCommand = "$sHomeDir/bin/fteCancelTransfer";
  $sCommandArgs = "-m $sAgentQM -p $sCoordinationQM -a " .
                  "$sFTEAgentName $sTransferId";
  @lsCmdout = (); # Truncate..

  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) {
    &logError ( 20091,
    "<%s> Insufficient time to proceed. Short by [%s] sec",
    $sMyName, abs $iTimeout );
    return 0; # Failure..
  }
  elsif ( WebSphereMQFTE::ExecLocalCmd ( $sFTEUser,
                                         $sEnvFile,
                                         $sCommand,
                                         $sCommandArgs,
                                         \$iExitCode,
                                         $iTimeout,
                                         \@lsCmdout ) ) {
    &logTrace ( 20092,
    "<%s> Cmd [%s] with args [%s] executed successfully. 
     And the transfer is cancelled successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logTrace ( 20093,
    "<%s> Cancel transfer for the Agent [%s] is not successful. 
     Check configuration",
    $sMyName, $sFTEAgentName );
  }

  return 0; # Failure..
} 
  
=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String FTEUser,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sFTEUser                  Unix user owner for WebSphere MQ FTE
 $sEnvFile                  Absolute path to environment file to source
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sFTEUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20094, "Subroutine <%s> called with:
  FTEUser                  [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sFTEUser, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20095,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20096,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20097,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20098,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20099,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  WebSphereMQFTE::ExecLocalCmd ( $sFTEUser, 
                                 $sEnvFile, 
                                 $sCmd, 
                                 $sCmdArgs, 
                                 \$iExitCode, 
                                 $iTimeout, 
                                 \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20100,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20101,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       # $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20102, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $riPid );                

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20103, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20104, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logTrace ( 20105, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20106,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20107,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20108,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( WebSphereMQFTE::GetPidDetails ( $iPid,
                                         \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20109,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20110,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20111,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20112,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20113, "<%s> PID [%s] does not exist",
    $sMyName, $iPid );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillWebSphereMQFTEProcesses ( IN String CoordinationQM,
                                               IN String DataDir,
                                               IN String FTEAgentName,
                                               IN String FTEUser );

 Description  : Kills processes for this instance of WebSphereMQFTE
 
 Parameters :

 $sCoordinationQM         Coordination Queue Manager of the WebSphere MQ FTE agent
 $sDataDir                Data Directory of the WebSphere MQ FTE
 $sFTEAgentName           Name of WebSphere MQ FTE agent
 $sFTEUser                Unix user owner for WebSphere MQ FTE
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillWebSphereMQFTEProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCoordinationQM,
       $sDataDir,
       $sFTEAgentName,
       $sFTEUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20114, "Subroutine <%s> called with:
  CoordinationQM           [%s]
  DataDir                  [%s]
  FTEAgentName             [%s]
  FTEUser                  [%s]",

  $sMyName, $sCoordinationQM, $sDataDir, $sFTEAgentName, $sFTEUser );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'WebSphereMQFTE::Start' ) { # During online..
    @liGoners = @liRunningProcs;                      # Get PIDs to kill..
  }
  else {
    WebSphereMQFTE::GetProcState ( \@liGoners,
                                   $sCoordinationQM,
                                   $sDataDir,
                                   $sFTEAgentName,
                                   $sFTEUser,
                                   0 ); # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20115,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20116,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20117,
    "<%s> Could not locate any process this instance of WebSphereMQFTE",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile       Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20118, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20119,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20120,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20121,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20122,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20123,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20124,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20125,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20126,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName:delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20127, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20128,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  if ( "solaris" eq $sSwitch ) {
      my $sOSVer = ( uname ) [2]; # Get OS version..
      if ($sOSVer =~ /\d+\.(\d+)/) {
          $sOSVer = $1;
      }
      $sPsCmd = "/usr/bin/ps";
      if ( $sOSVer < 11 ) {
          $sPsCmd = "/usr/ucb/ps" if ( -x "/usr/ucb/ps" );
      }

      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

  } elsif ( "hpux" eq $sSwitch ) {
      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  } elsif ( "aix" eq $sSwitch ) {
      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  } elsif ( "linux" eq $sSwitch ) {
      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }

  &logTrace ( 20129, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20130,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20131,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String CoordinationQM,
                                  IN String DataDir,
                                  IN String FTEAgentName,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sCoordinationQM           Coordination Queue Manager of the WebSphere MQ FTE agent
 $sDataDir                  Data Directory of the WebSphere MQ FTE
 $sFTEAgentName             Name of WebSphere MQ FTE agent
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCoordinationQM,
       $sDataDir,
       $sFTEAgentName,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20132, "Subroutine <%s> called with:
  CoordinationQM           [%s]
  DataDir                  [%s]
  FTEAgentName             [%s]
  SrvrDetails              [%s]",

  $sMyName,    $sCoordinationQM,     $sDataDir,     $sFTEAgentName,
  $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sLckFile, $sResourceName ) = ();

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "$sDataDir/config/$sCoordinationQM/agents/$sFTEAgentName";

  #------------------------------
  # Check PID folder exist or not
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20133, "<%s> PID or FTE Agent folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  else {
    &logTrace ( 20134, "<%s> PID or FTE Agent folder [%s] do not exist",
    $sMyName, $sPIDDir );
  }

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  $sProc = qr!java.*$sDataDir/(config)|(logs)/$sCoordinationQM/agents/$sFTEAgentName\b!;
  $sKey = $sFTEAgentName;
  $sPidFile = "$sPIDDir/agent.pid";
  $sLckFile = "$sPIDDir/agent.lck";
  if (! -e $sLckFile ) {
    $sPidFile = "$sDataDir/logs/$sCoordinationQM/agents/$sFTEAgentName/agent.pid";
    $sLckFile ="$sDataDir/logs/$sCoordinationQM/agents/$sFTEAgentName/agent.lck";
  }
  if (( -e $sPidFile) && ( -e $sLckFile)){
     $sLckFile = $sPidFile;
  }
  @{ $$rhlSrvrDetails { $sKey } } = ( $sLckFile, $sProc ); # Store..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20135, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
  
  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20195,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
  }
  else {
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    $sCommand = $sCmdHares; # Get global..
    ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
    $sCommandArgs = "-value $sResourceName $sAttribute";

    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                      =~ s/\n//;

    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20136,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iEngineTimeout ) { # Undefined value..
      &logTrace ( 20137, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20138, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }

    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    unless ( $iEngineTimeout ) {
      $sCommand = $sCmdHatype; # Get global..
      ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
      $sResourceType = "WebSphereMQFTE" if $sResourceType !~ /\w/; # Ensure..
      $sCommandArgs = "-value $sResourceType $sAttribute";

      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
  
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20139,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      elsif ( ! $iEngineTimeout ) { # Empty value..
        &logTrace ( 20140, "<%s> Could not get [%s] from engine",
        $sMyName, $sAttribute );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      else { # Success..
        &logTrace ( 20141, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
      }
    }
    else {
        &logTrace ( 20202, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20142,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

1; # So PERL knows its included..
